/* hangul.c -- This is a plug-in for the GIMP (Version 1.0's API).
 * Author: Chiduck Hwang <hwang@pseudo.snu.ac.kr>
 * modified from vftext.c
 * This plug-in uses FreeType engine for truetype font rendering.
 * FreeType Library can be found at 
 *    http://www.physiol.med.tu-muenchen.de/~robert/freetype.html or
 *    ftp://ftp.physiol.med.tu-muenchen.de/pub/freetype
 *
 * Time-stamp: <1997/07/24 22:02:08 narazaki@InetQ.or.jp>
 * Version: 0.66
 * Author: Shuji Narazaki <narazaki@InetQ.or.jp>
 * URL: http://www.inetq.or.jp/~narazaki/TheGIMP/
 *
 * To run this program, the following program (library) should be installed:
 *
 *  VFlib (c) Hirotsugu KAKUGAWA (kakugawa@se.hiroshima-u.ac.jp)
 *    URL: ftp://gull.se.hiroshima-u.ac.jp/pub/VFlib/
 *
 * These prorgrams are distrbuted under GNU GPL.
 * And this program also must be distributed under the following condition:
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Compile:
 */

#include <gtk/gtk.h>
#include <libgimp/gimp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
/* header from FreeType Project */
#include "freetype.h"

/* parameters */
#define	PLUG_IN_NAME	"plug_in_freetype"
#define SHORT_NAME	"freetype"
#if HANGUL
#define CONFIG_FILE "/etc/hangul4gimp.conf"
#define	MENU_POSITION	"<Image>/Filters/Render/Ÿ"
#define PROGRESS_NAME	"۾  ..."
#else
#define CONFIG_FILE "/etc/truetype4gimp.conf"
#define	MENU_POSITION	"<Image>/Filters/Render/FreeType"
#define PROGRESS_NAME	"Rendering FreeType..."
#endif

#define VALS		hangulvals
#define INTERFACE	hangul_interface
#define	DIALOG		hangul_dialog
#define ERROR_DIALOG	hangul_error_dialog
#define	OK_CALLBACK	hangul_ok_callback
#define MAX_LINE	100

#define UNI_ENCODING 0
#if HANGUL
#define WANSUNG_ENCODING 1
#endif

void *memset();
void bzero();
static void query(void);
static void run(char *name,
				int nparams,
				GParam * param,
				int *nreturn_vals,
				GParam ** return_vals);
static GStatusType hangul_render(gint32 image_id,
								 gint32 layer_id,
								 gint32 * new_layer,
								 guchar * text);
static int hangul_get_char_image(guint16 kscode);
static int hangul_render_to_image(gint32 image_id,
								  gint32 * layer_id,
								  gint32 beneath_layer,
								  guchar * text);
void set_pixel_color(guchar * dest,
					 int x,
					 int y,
					 int width,
					 guchar intensity,
					 int ch,
					 guchar * color);

static int DIALOG();
static void ERROR_DIALOG(gint gtk_was_not_initialized, guchar * message);
static void OK_CALLBACK(GtkWidget * widget, gpointer data);
static void hangul_open_fontname_dialog(GtkWidget * widget, gpointer data);
static void hangul_fontname_callback(GtkWidget * widget, gpointer data);

/* gtkWrapper functions */
#define PROGRESS_UPDATE_NUM	50
#define ENTRY_WIDTH	120
#define SCALE_WIDTH	100
static void gtkW_close_callback(GtkWidget * widget, gpointer data);
static void gtkW_gint_update(GtkWidget * widget, gpointer data);
static void gtkW_toggle_update(GtkWidget * widget, gpointer data);
static void gtkW_string_update(GtkWidget * widget, gpointer data);
static GtkWidget *gtkW_dialog_new(char *name,
								  GtkSignalFunc ok_callback,
								  GtkSignalFunc close_callback);
static GtkWidget *gtkW_error_dialog_new(char *name);
static void gtkW_table_add_toggle(GtkWidget * table,
								  gchar * name,
								  gint x1,
								  gint x2,
								  gint y,
								  GtkSignalFunc update,
								  gint * value);
static void gtkW_table_add_gint(GtkWidget * table,
								gchar * name,
								gint x,
								gint y,
								GtkSignalFunc update,
								gint * value,
								gchar * buffer);
GtkWidget *gtkW_check_button_new(GtkWidget * parent,
								 gchar * name,
								 GtkSignalFunc update,
								 gint * value);
GtkWidget *gtkW_frame_new(GtkWidget * parent, gchar * name);
GtkWidget *gtkW_table_new(GtkWidget * parent, gint col, gint row);
GtkWidget *gtkW_hbox_new(GtkWidget * parent);
GtkWidget *gtkW_vbox_new(GtkWidget * parent);

GPlugInInfo PLUG_IN_INFO =
{
	NULL,						/* init_proc  */
	NULL,						/* quit_proc */
	query,						/* query_proc */
	run,						/* run_proc */
};

typedef struct {
	guchar fontname[64];
	gint32 pointsize;
	gint32 tategaki;
	gint32 border;
	gint32 spacing;				/* between letters */
	gint32 antialiasing;
	gfloat x;
	gfloat y;
	guchar text[256];
	gint32 h_stretch;
} HangulValues;

static HangulValues VALS =
{
#if HANGUL
	"batang",						/* font name */
#else
	"arial",
#endif
	64,							/* font size */
	0,							/* encoding */
	10,							/* border */
	2,							/* spacing */
	1,							/* antialiasing */
	0,							/* x position */
	0,							/* y position */
#if HANGUL
	"ѱۻ",					/* text to draw */
#else
	"Gimp",
#endif
	100
};

typedef struct {
	gint run;
} HangulInterface;

static HangulInterface INTERFACE =
{FALSE};

GtkWidget *fontname_entry;
GRunModeType run_mode;

/* variables */
static int WinWidth, WinHeight;
/* supersampling parameter */
static int xdelta, ydelta;
static guchar *text = NULL;


/* variables for FreeType library */
static TT_UShort upm;
static TT_CharMap cmap;
static TT_Raster_Map raster;
static TT_Glyph glyph;
static TT_Glyph_Metrics metrics;
#if 0
static TT_Glyph_Outline outline;
#endif
static TT_Face face;
static TT_Face_Properties properties;
static TT_Instance_Metrics imetrics;
static TT_Instance instance;
static int encoding;
static int font_index = 0;
static int hangul_font_width = -1;
static int hangul_font_height = -1;
TT_Short pid, eid;

int flags = TTLOAD_DEFAULT;
int xoffset = 0, yoffset = 0;
int nocmap;
#if HANGUL
int is_hangul_font = 1;
#else
int is_hangul_font = 0;
#endif

/* the routines  and definition for FreeType library */
#define PIXEL(x)   (((x)+(upm/2))/upm)
char * get_font_filename(char *fontname);
int hangul_open_font( HangulValues *val);
guint16 get_glyph_index(guint16 code);
void get_cmap(void);
char ** get_font_list(int *num);
int get_string_width(guchar *);



/* string to KS code */
#define HangulConv(ch1, ch2) ((ch1)*0x100 + (ch2))

MAIN();

static void query()
{
	static GParamDef args[] =
	{
		{PARAM_INT32, "run_mode", "Interactive, non-interactive"},
		{PARAM_IMAGE, "image", "Input image"},
		{PARAM_DRAWABLE, "drawable", "Input drawable (not used)"},
		{PARAM_FLOAT, "x", "the x coordinate for the left side of text bounding box"},
		{PARAM_FLOAT, "y", "the y coordinate for the left side of text bounding box"},
		{PARAM_STRING, "text", "hangul characters. "},
		{PARAM_FLOAT, "h_stretch", "how much we stretch horizontally"},
		{PARAM_INT32, "border", "the size of the border: border >= 0"},
		{PARAM_INT32, "antialias", "generate antialiased text"},
	{PARAM_FLOAT, "size", "the size of text in either pixels or points"},
		{PARAM_INT32, "size_type", "not used"},
		{PARAM_STRING, "foundry", "the font foundry, (currently ignored)"},
		{PARAM_STRING, "family", "the font family, (currently ignored)"},
		{PARAM_STRING, "weight", "the font weight, (currently ignored)"},
		{PARAM_STRING, "slant", "the font slant, (currently ignored)"},
		{PARAM_STRING, "set_width", "the font set-width parameter, (currently ignored)"},
	{PARAM_INT32, "spacing", "the font spacing (NOT STRING BUT INTEGER)"}
	};
	static GParamDef return_vals[] =
	{
		{PARAM_DRAWABLE, "text_layer", "new hangul text layer"}
	};
	static int nargs = sizeof(args) / sizeof(args[0]);
	static int nreturn_vals = sizeof(return_vals) / sizeof(return_vals[0]);;

	gimp_install_procedure(PLUG_IN_NAME,
#if HANGUL
						   "ѱ ϴ  ̾ ϴ",
						   "ѱ ϴ  ̾ ̹ ߰ŵϴ.",
						   "Ȳ ġ(hwang@pseudo.snu.ac.kr)",
						   "Ȳ ġ",
#else
						   "Add a new layer for texts rendered by FreeType engine",
						   "Add a new layer to the image. The layer holds letters.",
						   "Chiduck Hwang(hwang@pseudo.snu.ac.kr)",
						   "Chiduck Hwang",
#endif
						   "1997",
						   MENU_POSITION,
						   "RGB*, GRAY*",
						   PROC_PLUG_IN,
						   nargs, nreturn_vals,
						   args, return_vals);
}

static void run(char *name,
				int nparams,
				GParam * param,
				int *nreturn_vals,
				GParam ** return_vals)
{
	static GParam values[1];
	GStatusType status = STATUS_EXECUTION_ERROR;
	gint32 new_layer;

	run_mode = param[0].data.d_int32;

	*nreturn_vals = 2;
	*return_vals = values;

	values[0].type = PARAM_STATUS;
	values[0].data.d_status = status;

	switch (run_mode) {
	case RUN_INTERACTIVE:
		{
			char **fontname = get_font_list(NULL);
			if(fontname) {
				strcpy(VALS.fontname, fontname[0]);
			}
		}
		gimp_get_data(PLUG_IN_NAME, &VALS);
		if (!DIALOG())
			return;
		text = VALS.text;
		if ((text == NULL) || (strlen(text) == 0)) {
			char buffer[1024];

#if HANGUL
			sprintf(buffer, "ڿ ϴ.  ڿ ϼ.");
#else
			sprintf(buffer, "Error: text is empty!!.");
#endif
			ERROR_DIALOG(0, buffer);
			return;
		}
		break;
	case RUN_NONINTERACTIVE:
		/* build up the val structure from the parameters */
		VALS.x = param[3].data.d_float;
		VALS.y = param[4].data.d_float;
		text = param[5].data.d_string;
		VALS.h_stretch = (gint)param[6].data.d_float;
		VALS.border = param[7].data.d_int32;
		VALS.antialiasing = param[8].data.d_int32;
		VALS.pointsize = (gint) param[9].data.d_float;
		strcpy(VALS.fontname, param[12].data.d_string);
		VALS.spacing = param[16].data.d_int32;
		break;
	case RUN_WITH_LAST_VALS:
		gimp_get_data(PLUG_IN_NAME, &VALS);
		text = VALS.text;
		break;
	}

	status = hangul_render(param[1].data.d_image, param[2].data.d_drawable,
						   &new_layer, text);

	if (run_mode != RUN_NONINTERACTIVE)
		gimp_displays_flush();
	if (run_mode == RUN_INTERACTIVE && status == STATUS_SUCCESS)
		gimp_set_data(PLUG_IN_NAME, &VALS, sizeof(HangulValues));

	values[0].type = PARAM_STATUS;
	values[0].data.d_status = status;

	if (status == STATUS_SUCCESS) {
		values[1].type = PARAM_DRAWABLE;
		values[1].data.d_int32 = new_layer;
	}
}

static GStatusType
hangul_render(gint32 image_id,
			   gint32 beneath_layer,
			   gint32 * new_layer,
			   guchar * text)
{

	if (TT_Init_FreeType()) {
		if (run_mode == RUN_INTERACTIVE)
			ERROR_DIALOG(0, "Fail to initializing FreeType Library");
		return STATUS_EXECUTION_ERROR;
	}
	if (hangul_open_font(&VALS)) {
		if (run_mode == RUN_INTERACTIVE)
#if HANGUL
			ERROR_DIALOG(0, "Ͻ ۲  ϴ.");
#else
			ERROR_DIALOG(0, "Fail to open for the specified font");
#endif
		return STATUS_EXECUTION_ERROR;
	}
	if(is_hangul_font) {
		TT_Load_Glyph(instance, glyph, get_glyph_index(0xb3b3), flags);
	} else {
		TT_Load_Glyph(instance, glyph, get_glyph_index('L'), flags);
	}
	TT_Get_Glyph_Metrics(glyph, &metrics);
	hangul_font_width = metrics.advanceWidth>>6;

	raster.flow = TT_Flow_Down;
	if(is_hangul_font) {
		raster.width = hangul_font_width;
	} else {
		raster.width = hangul_font_width * 2;
	}
	raster.rows = hangul_font_height;
	raster.cols = (raster.width + 7) >> 3;
	raster.size = raster.cols * raster.rows;
	raster.bitmap = g_new(char, raster.size);

	if(strncmp(text, "< ", 2) == 0) {
		int fd;
		char *filename = text + 2;
		while(isspace(*filename)) filename++;
		if((fd = open(filename, O_RDONLY)) > 0) {
			struct stat statbuf;
			char *buf;
			fstat(fd, &statbuf);
			buf = g_new(char, statbuf.st_size);
			if(read(fd, text, statbuf.st_size) != statbuf.st_size) {
				g_free(buf);
			} else 
				text = buf;
			close(fd);
		}
	} else if(strstr(text, "\\n") != NULL) {
		guchar *s = text;
		while(*s) {
			if(*s == '\\') {
				if(*(s + 1) == 'n') {
					*s = '\n';
					memmove(s + 1, s + 2, strlen(s+1));
				} else if(*(s + 1)) s++;
			} 
			s++;
		}
	}
	hangul_render_to_image(image_id, new_layer, beneath_layer, text);

	TT_Close_Face(face);
	TT_Done_FreeType();

	return STATUS_SUCCESS;
}

static int
hangul_get_char_image(guint16 kscode)
{
	int idx;

	bzero(raster.bitmap, raster.size);
	idx = get_glyph_index(kscode);
	TT_Load_Glyph(instance, glyph, idx, flags);
	TT_Get_Glyph_Metrics(glyph, &metrics);
	TT_Get_Glyph_Bitmap(glyph, &raster, xoffset << 6, yoffset << 6);
	return metrics.advanceWidth>>6;
}

int hangul_render_to_image(gint32 image_id,
						   gint32 * layer_id,
						   gint32 beneath_layer,
						   guchar * text)
{
	GPixelRgn pixel_rgn;
	GDrawable *drawable;
	guchar fore[4];				/* rgb + alpha */
	guchar *ylbuff[MAX_LINE];
	guchar *dest = NULL;
	gint xshift, yshift, slen, i, x, y, n, intensity;
	gint pixel_count = 0, total_pixel_num;
	gint index, line_num, last_eol;
	gint len, ch_num;
	guchar *buf;
	buf = (guchar *)raster.bitmap;

	for (index = line_num = last_eol = 0; index < strlen(text); index++)
		if (text[index] == '\n') {
			len = index - last_eol + 1;
			ylbuff[line_num] = g_new(guchar, len);
			memcpy(ylbuff[line_num], text + last_eol, len - 1);
			ylbuff[line_num][len - 1] = '\000';		/* newline -> terminator */
			last_eol = index + 1;
			if (MAX_LINE <= ++line_num)
				break;
		}
	if (last_eol < index) {		/* reached eof without newline */
		len = index - last_eol + 1;
		ylbuff[line_num] = g_new(guchar, len);
		memcpy(ylbuff[line_num], text + last_eol, len - 1);
		ylbuff[line_num++][len - 1] = '\000';	/* newline -> terminator */
	}

	if (VALS.tategaki) {
		WinWidth = 2 * VALS.border + hangul_font_width * line_num + VALS.spacing * (line_num - 1);
		WinHeight = 0;
		for(i=0;i<line_num;i++) {
			int height = hangul_font_height * strlen(ylbuff[i]) + VALS.spacing * (strlen(ylbuff[i]) - 1);
			if(height > WinHeight) WinHeight = height;
		}
		WinHeight += 2 * VALS.border;
	} else {
		WinHeight = 2 * VALS.border + hangul_font_height/ ydelta * line_num + VALS.spacing * (line_num - 1);
		WinWidth = 0;
		for(i=0;i<line_num;i++) {
			int width = get_string_width(ylbuff[i]);
			if(WinWidth < width) WinWidth = width;
		}
		WinWidth += 2 * VALS.border;
	}

	total_pixel_num = WinWidth * WinHeight;

	if (gimp_image_base_type(image_id) == RGB) {
#if HANGUL
		*layer_id = gimp_layer_new(image_id, "ѱ", WinWidth, WinHeight,
#else
		*layer_id = gimp_layer_new(image_id, "text", WinWidth, WinHeight,
#endif
								   RGBA_IMAGE, 100, NORMAL_MODE);
		ch_num = 4;				/* rgb + alpha */
	} else {
#if HANGUL
		*layer_id = gimp_layer_new(image_id, "ѱ", WinWidth, WinHeight,
#else
		*layer_id = gimp_layer_new(image_id, "text", WinWidth, WinHeight,
#endif
								   GRAYA_IMAGE, 100, NORMAL_MODE);
		ch_num = 2;				/* value + alpha */
	}

	gimp_layer_set_preserve_transparency(*layer_id, 0);
	drawable = gimp_drawable_get(*layer_id);

	gimp_pixel_rgn_init(&pixel_rgn, drawable, 0, 0, WinWidth, WinHeight, TRUE, TRUE);
	/* clear the new layer (make transparent) */
	{
		gpointer pr;
		gint x, y;

		pr = gimp_pixel_rgns_register(1, &pixel_rgn);

		for (; pr != NULL; pr = gimp_pixel_rgns_process(pr)) {
			guchar *src = pixel_rgn.data;

			for (y = 0; y < pixel_rgn.h; y++) {
				guchar *s = src;
				for (x = 0; x < pixel_rgn.w; x++) {
					gint b;

					for (b = 0; b < ch_num; b++)
						*(s + b) = 0;
					s += pixel_rgn.bpp;
				}
				src += pixel_rgn.rowstride;
			}
		}
	}
	/* use red value if the layer is grayscale */
	gimp_palette_get_foreground(fore + 0, fore + 1, fore + 2);
	fore[3] = 255;

	if (VALS.tategaki) {
		xshift = WinWidth - VALS.border - hangul_font_width/xdelta;
		yshift = VALS.border;
	} else {
		xshift = VALS.border;
		yshift = VALS.border;
	}

	gimp_progress_init(PROGRESS_NAME);
	dest = g_new(guchar , WinWidth * drawable->bpp);

	for (n = 0; n < line_num; n++) {	/* a line */
		slen = strlen(ylbuff[n]);
		for (i = 0; i < slen; ) {		/* a letter */
			int advance_width;
			bzero(dest, WinWidth * drawable->bpp);
			if(is_hangul_font
#if HANGUL
			   && (ylbuff[n][i] & 0x80)
#endif
			   ) {
				advance_width = hangul_get_char_image(HangulConv(ylbuff[n][i], ylbuff[n][i + 1]));
				i += 2;
			} else {
				advance_width = hangul_get_char_image(ylbuff[n][i]);
				i++;
			}
			for (y = 0; y < hangul_font_height; y += ydelta) {
				for (x = 0; x < advance_width/xdelta; x++) {
					if(VALS.antialiasing) {
						int k,l;
						intensity = 0;
						for(l = 0; l< 8;l++) {
							int c = buf[(y+l)*raster.cols + x];
							for(k=0;k < 8;k++) {
								if(c & 1) intensity++;
								c >>= 1;
							}
						}
						intensity = intensity * 255.0 /64;
					} else {
						intensity = buf[y * raster.cols + (x>>3)] & (1<<(7-(x&7)));
						if(intensity) intensity = 255;
						else intensity = 0;
					}
					set_pixel_color(dest, x, 0, WinWidth, (guchar) intensity,
									ch_num, fore);
					if (!(++pixel_count % (total_pixel_num / PROGRESS_UPDATE_NUM)))
						gimp_progress_update((double) pixel_count / total_pixel_num);
				}
				gimp_pixel_rgn_set_row(&pixel_rgn, dest, xshift, yshift + y/ydelta, advance_width/xdelta);
			}
			if (VALS.tategaki)
				yshift += hangul_font_height/ydelta + VALS.spacing;
			else
				xshift += advance_width/xdelta + VALS.spacing;
		}
		if (VALS.tategaki) {
			xshift -= hangul_font_width/xdelta + VALS.spacing;
			yshift = VALS.border;
		} else {
			xshift = VALS.border;
			yshift += hangul_font_height/ydelta + VALS.spacing;
		}
	}
	gimp_progress_update(1.0);
	gimp_drawable_flush(drawable);
	gimp_layer_set_offsets(drawable->id, VALS.x, VALS.y);
	gimp_drawable_merge_shadow(drawable->id, TRUE);
	gimp_drawable_update(drawable->id, 0, 0, WinWidth, WinHeight);
	gimp_image_add_layer(image_id, *layer_id, -1);
	gimp_drawable_detach(drawable);
	return 1;
}

void set_pixel_color(guchar * dest,
					 int x,
					 int y,
					 int width,
					 guchar intensity,
					 int ch,
					 guchar * color)
{
	int index;
	guchar *base;

	base = dest + (width * y + x) * ch;
	if (!VALS.antialiasing)
		intensity = (intensity > 127) ? 255 : 0;
	for (index = 0; index < ch; index++)
		*base++ = *color++;
	*(base - 1) = intensity;
}


/* Hangul interface functions  */
static int DIALOG()
{
	GtkWidget *dlg;
	GtkWidget *vbox;
	GtkWidget *entry;
	GtkWidget *frame;
	GtkWidget *table;
	gchar **argv;
	gint argc;
	static gchar buffer[4][32];

	argc = 1;
	argv = g_new(gchar *, 1);
	argv[0] = g_strdup(PLUG_IN_NAME);

	gtk_init(&argc, &argv);
	gtk_rc_parse(gimp_gtkrc());

	dlg = gtkW_dialog_new(PLUG_IN_NAME,
						  (GtkSignalFunc) OK_CALLBACK,
						  (GtkSignalFunc) gtkW_close_callback);
	vbox = gtkW_vbox_new(GTK_DIALOG(dlg)->vbox);

	/* file name entry */
	frame = gtkW_frame_new(vbox, "ڿ");
	entry = gtk_entry_new();
	gtk_widget_set_usize(entry, ENTRY_WIDTH, 0);
	gtk_entry_set_text(GTK_ENTRY(entry), VALS.text);
	gtk_signal_connect(GTK_OBJECT(entry), "changed",
					   (GtkSignalFunc) gtkW_string_update,
					   &VALS.text);
	gtk_container_add(GTK_CONTAINER(frame), entry);
	gtk_widget_show(entry);

	/* parameter settings */
	frame = gtkW_frame_new(vbox, "Parameter Settings");
	table = gtkW_table_new(frame, 8, 2);	/* 4 raw, 2 columns(name and value) */

#if HANGUL
	gtkW_table_add_toggle(table, " ", 0, 2, 1,
#else
	gtkW_table_add_toggle(table, "no effect for non-korean", 0, 2, 1,
#endif
						  (GtkSignalFunc) gtkW_toggle_update,
						  &VALS.tategaki);
	gtkW_table_add_toggle(table, "Antialiasing", 0, 2, 2,
						  (GtkSignalFunc) gtkW_toggle_update,
						  &VALS.antialiasing);
	{
		GtkWidget *button;

		button = gtk_button_new_with_label("Font Name");
		gtk_signal_connect(GTK_OBJECT(button), "clicked",
					   (GtkSignalFunc) hangul_open_fontname_dialog, dlg);
		gtk_table_attach(GTK_TABLE(table), button, 0, 1, 3, 4,
						 GTK_FILL | GTK_EXPAND, GTK_FILL, 5, 0);
		gtk_widget_show(button);

		fontname_entry = gtk_entry_new();
		gtk_widget_set_usize(GTK_WIDGET(fontname_entry), ENTRY_WIDTH, 0);
		gtk_entry_set_text(GTK_ENTRY(fontname_entry), VALS.fontname);
		gtk_signal_connect(GTK_OBJECT(fontname_entry), "changed",
						   (GtkSignalFunc) gtkW_string_update,
						   &VALS.fontname);
		gtk_table_attach(GTK_TABLE(table), fontname_entry, 1, 2, 3, 4,
					GTK_FILL | GTK_EXPAND | GTK_SHRINK, GTK_FILL, 10, 0);
		gtk_widget_show(fontname_entry);
	}

#if HANGUL
	gtkW_table_add_gint(table, " ũ(point)", 0, 4,
#else
	gtkW_table_add_gint(table, "font size(point)", 0, 4,
#endif
						(GtkSignalFunc) gtkW_gint_update,
						&VALS.pointsize, buffer[0]);
#if HANGUL
	gtkW_table_add_gint(table, "ѷ ", 0, 5,
#else
	gtkW_table_add_gint(table, "border", 0, 5,
#endif
						(GtkSignalFunc) gtkW_gint_update,
						&VALS.border, buffer[1]);
#if HANGUL
	gtkW_table_add_gint(table, "ڰ ", 0, 6,
#else
	gtkW_table_add_gint(table, "spacing", 0, 6,
#endif
						(GtkSignalFunc) gtkW_gint_update,
						&VALS.spacing, buffer[2]);
#if HANGUL
	gtkW_table_add_gint(table, " (%)", 0, 7,
#else
	gtkW_table_add_gint(table, "horizontal stretch(%)", 0, 7,
#endif
						(GtkSignalFunc) gtkW_gint_update,
						&VALS.h_stretch, buffer[3]);

	gtk_widget_show(dlg);

	gtk_main();
	gdk_flush();

	return INTERFACE.run;
}

static void hangul_open_fontname_dialog(GtkWidget * widget, gpointer data)
{
	/* The logo frame & drawing area */
	GtkWidget *dlg, *vbox, *frame, *button;

	dlg = gtk_dialog_new();
	gtk_window_set_title(GTK_WINDOW(dlg), "Font Selector");
	gtk_window_position(GTK_WINDOW(dlg), GTK_WIN_POS_MOUSE);
	gtk_signal_connect(GTK_OBJECT(dlg), "destroy",
					   (GtkSignalFunc) gtkW_close_callback, NULL);
	/* Action Area */
	button = gtk_button_new_with_label("OK");
	GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
	gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
							  (GtkSignalFunc) gtk_widget_destroy,
							  GTK_OBJECT(dlg));
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg)->action_area), button,
					   TRUE, TRUE, 0);
	gtk_widget_grab_default(button);
	gtk_widget_show(button);

	button = gtk_button_new_with_label("Cancel");
	GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
	gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
							  (GtkSignalFunc) gtk_widget_destroy,
							  GTK_OBJECT(dlg));
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg)->action_area), button,
					   TRUE, TRUE, 0);
	gtk_widget_show(button);

	vbox = gtkW_vbox_new(GTK_DIALOG(dlg)->vbox);
	frame = gtkW_frame_new(vbox, "Font Name");

	{
		GtkWidget *scrolled_win;
		GtkWidget *list, *list_item;
		int i;
		int nlist_items;
		char **fontnames;

		fontnames = get_font_list(&nlist_items);

		scrolled_win = gtk_scrolled_window_new(NULL, NULL);
		gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_win),
							 GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
		gtk_container_add(GTK_CONTAINER(frame), scrolled_win);
		list = gtk_list_new();
		gtk_list_set_selection_mode(GTK_LIST(list), GTK_SELECTION_BROWSE);
		gtk_container_add(GTK_CONTAINER(scrolled_win), list);
		gtk_widget_show(list);

		for (i = 0; i < nlist_items; i++) {
			list_item = gtk_list_item_new_with_label(fontnames[i]);
			gtk_container_add(GTK_CONTAINER(list), list_item);
			gtk_signal_connect(GTK_OBJECT(list_item), "select",
							   (GtkSignalFunc) hangul_fontname_callback,
							   fontnames[i]);
			gtk_widget_show(list_item);
			if(strcmp(fontnames[i], VALS.fontname) == 0) {
				gtk_list_select_item(GTK_LIST(list), i);
			}
		}
		gtk_widget_set_usize(GTK_WIDGET(scrolled_win), 300, 200);
		gtk_widget_show(scrolled_win);
	}
	gtk_widget_show(dlg);

	gtk_main();
	gdk_flush();
}

static void hangul_fontname_callback(GtkWidget * widget, gpointer data)
{
	int i;

	/*  Is this font being selected?  */
	if (widget->state != GTK_STATE_SELECTED)
		return;

	strcpy(VALS.fontname, data);
	for (i = 0; i < strlen(VALS.fontname); i++)
		if (VALS.fontname[i] == ' ') {
			VALS.fontname[i] = '\000';
			break;
		}
	if (!INTERFACE.run)
		gtk_entry_set_text(GTK_ENTRY(fontname_entry), VALS.fontname);
	gtk_widget_draw(fontname_entry, NULL);
}

static void ERROR_DIALOG(gint gtk_was_not_initialized, guchar * message)
{
	GtkWidget *dlg;
	GtkWidget *table;
	GtkWidget *label;
	gchar **argv;
	gint argc;

	if (gtk_was_not_initialized) {
		argc = 1;
		argv = g_new(gchar *, 1);
		argv[0] = g_strdup(PLUG_IN_NAME);
		gtk_init(&argc, &argv);
		gtk_rc_parse(gimp_gtkrc());
	}
	dlg = gtkW_error_dialog_new(PLUG_IN_NAME);

	table = gtk_table_new(1, 1, FALSE);
	gtk_container_border_width(GTK_CONTAINER(table), 10);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg)->vbox), table, TRUE, TRUE, 0);

	label = gtk_label_new(message);
	gtk_table_attach(GTK_TABLE(table), label, 0, 1, 0, 1, GTK_FILL | GTK_EXPAND,
					 0, 0, 0);
	gtk_widget_show(label);
	gtk_widget_show(table);

	gtk_main();
	gdk_flush();
}

static void OK_CALLBACK(GtkWidget * widget,
						gpointer data)
{
	INTERFACE.run = TRUE;
	gtk_widget_destroy(GTK_WIDGET(data));
}

/* Hangul interface functions  */
static void gtkW_gint_update(GtkWidget * widget,
							 gpointer data)
{
	*(gint *) data = (gint) atof(gtk_entry_get_text(GTK_ENTRY(widget)));
}

static void gtkW_close_callback(GtkWidget * widget,
								gpointer data)
{
	gtk_main_quit();
}

static void gtkW_toggle_update(GtkWidget * widget,
							   gpointer data)
{
	int *toggle_val;

	toggle_val = (int *) data;

	if (GTK_TOGGLE_BUTTON(widget)->active)
		*toggle_val = TRUE;
	else
		*toggle_val = FALSE;
}

/* gtkW is the abbreviation of gtk Wrapper */
static GtkWidget *
 gtkW_dialog_new(char *name,
				 GtkSignalFunc ok_callback,
				 GtkSignalFunc close_callback)
{
	GtkWidget *dlg, *button;

	dlg = gtk_dialog_new();
	gtk_window_set_title(GTK_WINDOW(dlg), name);
	gtk_window_position(GTK_WINDOW(dlg), GTK_WIN_POS_MOUSE);
	gtk_signal_connect(GTK_OBJECT(dlg), "destroy",
					   (GtkSignalFunc) gtkW_close_callback, NULL);

	/* Action Area */
	button = gtk_button_new_with_label("OK");
	GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
					   (GtkSignalFunc) ok_callback, dlg);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg)->action_area), button,
					   TRUE, TRUE, 0);
	gtk_widget_grab_default(button);
	gtk_widget_show(button);

	button = gtk_button_new_with_label("Cancel");
	GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
	gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
							  (GtkSignalFunc) gtk_widget_destroy,
							  GTK_OBJECT(dlg));
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg)->action_area), button,
					   TRUE, TRUE, 0);
	gtk_widget_show(button);

	return dlg;
}

static GtkWidget *
 gtkW_error_dialog_new(char *name)
{
	GtkWidget *dlg, *button;

	dlg = gtk_dialog_new();
	gtk_window_set_title(GTK_WINDOW(dlg), name);
	gtk_window_position(GTK_WINDOW(dlg), GTK_WIN_POS_MOUSE);
	gtk_signal_connect(GTK_OBJECT(dlg), "destroy",
					   (GtkSignalFunc) gtkW_close_callback, NULL);

	/* Action Area */
	button = gtk_button_new_with_label("OK");
	GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
					   (GtkSignalFunc) gtkW_close_callback, dlg);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg)->action_area), button,
					   TRUE, TRUE, 0);
	gtk_widget_grab_default(button);
	gtk_widget_show(button);
	gtk_widget_show(dlg);

	return dlg;
}

GtkWidget *
 gtkW_table_new(GtkWidget * parent, gint col, gint row)
{
	GtkWidget *table;

	table = gtk_table_new(col, row, FALSE);
	gtk_container_border_width(GTK_CONTAINER(table), 10);
	gtk_container_add(GTK_CONTAINER(parent), table);
	gtk_widget_show(table);

	return table;
}

GtkWidget *
 gtkW_hbox_new(GtkWidget * parent)
{
	GtkWidget *hbox;

	hbox = gtk_hbox_new(FALSE, 5);
	gtk_container_border_width(GTK_CONTAINER(hbox), 5);
	gtk_box_pack_start(GTK_BOX(parent), hbox, FALSE, TRUE, 0);
	gtk_widget_show(hbox);

	return hbox;
}

GtkWidget *
 gtkW_vbox_new(GtkWidget * parent)
{
	GtkWidget *vbox;

	vbox = gtk_vbox_new(FALSE, 5);
	gtk_container_border_width(GTK_CONTAINER(vbox), 10);
	/* gtk_box_pack_start (GTK_BOX (parent), vbox, TRUE, TRUE, 0); */
	gtk_container_add(GTK_CONTAINER(parent), vbox);
	gtk_widget_show(vbox);

	return vbox;
}

GtkWidget *
 gtkW_check_button_new(GtkWidget * parent,
					   gchar * name,
					   GtkSignalFunc update,
					   gint * value)
{
	GtkWidget *toggle;

	toggle = gtk_check_button_new_with_label(name);
	gtk_signal_connect(GTK_OBJECT(toggle), "toggled",
					   (GtkSignalFunc) update,
					   value);
	gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(toggle), *value);
	gtk_container_add(GTK_CONTAINER(parent), toggle);
	gtk_widget_show(toggle);
	return toggle;
}

GtkWidget *
 gtkW_frame_new(GtkWidget * parent,
				gchar * name)
{
	GtkWidget *frame;

	frame = gtk_frame_new(name);
	gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);
	gtk_container_border_width(GTK_CONTAINER(frame), 5);
	gtk_box_pack_start(GTK_BOX(parent), frame, FALSE, FALSE, 0);
	gtk_widget_show(frame);
	return frame;
}

static void gtkW_table_add_toggle(GtkWidget * table,
								  gchar * name,
								  gint x1,
								  gint x2,
								  gint y,
								  GtkSignalFunc update,
								  gint * value)
{
	GtkWidget *toggle;

	toggle = gtk_check_button_new_with_label(name);
	gtk_table_attach(GTK_TABLE(table), toggle, x1, x2, y, y + 1,
					 GTK_FILL | GTK_EXPAND, 0, 0, 0);
	gtk_signal_connect(GTK_OBJECT(toggle), "toggled",
					   (GtkSignalFunc) update,
					   value);
	gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(toggle), *value);
	gtk_widget_show(toggle);
}

static void gtkW_table_add_gint(GtkWidget * table,
								gchar * name,
								gint x,
								gint y,
								GtkSignalFunc update,
								gint * value,
								gchar * buffer)
{
	GtkWidget *label, *entry;

	label = gtk_label_new(name);
	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
	gtk_table_attach(GTK_TABLE(table), label, x, x + 1, y, y + 1,
					 GTK_FILL | GTK_EXPAND, GTK_FILL, 5, 0);
	gtk_widget_show(label);

	entry = gtk_entry_new();
	gtk_table_attach(GTK_TABLE(table), entry, x + 1, x + 2, y, y + 1,
					 GTK_FILL | GTK_EXPAND, GTK_FILL, 10, 0);
	gtk_widget_set_usize(entry, ENTRY_WIDTH, 0);
	sprintf(buffer, "%d", *(gint *) value);
	gtk_entry_set_text(GTK_ENTRY(entry), buffer);
	gtk_signal_connect(GTK_OBJECT(entry), "changed",
					   (GtkSignalFunc) update, value);
	gtk_widget_show(entry);
}

static void gtkW_string_update(GtkWidget * widget,
							   gpointer data)
{
	guchar *new_string;

	new_string = gtk_entry_get_text(GTK_ENTRY(widget));
	strcpy(data, new_string);
}

int
hangul_open_font( HangulValues *val)
{
	char *font_filename;
	float x_res, y_res;

	font_filename = get_font_filename(val->fontname);
	if(font_filename == NULL) {
		return 1;
	}
	if(font_index <= 0) {
		if(TT_Open_Face(font_filename, &face)) return 1;
	} else {
		if(TT_Open_Collection(font_filename, font_index, &face)) return 1;
	}
	get_cmap();
	if(TT_Get_Face_Properties(face, &properties)) {
		return 1;
	}
	upm = properties.header->Units_Per_EM;

	if(TT_New_Instance(face, &instance)) return 1;
	x_res = VALS.h_stretch * 72.27 / 100;
	y_res = 72.27;
	if(VALS.antialiasing) {
		xdelta = ydelta = 8;
		x_res *= 8;
		y_res *= 8;
	} else {
		xdelta = ydelta = 1;
	}
	if(TT_Set_Instance_Resolution(instance, x_res, y_res)) return 1;
	if(TT_Set_Instance_PointSize(instance, val->pointsize)) return 1;
	if(TT_Get_Instance_Metrics(instance, &imetrics)) return 1;

	if(TT_New_Glyph(face, &glyph)) return 1;

	yoffset = - PIXEL(properties.header->yMin * imetrics.y_ppem);

	flags = TTLOAD_DEFAULT;
	hangul_font_height = imetrics.y_ppem;
	if(VALS.antialiasing) {
		hangul_font_height = (hangul_font_height/8)*8;
	}

	return 0;
}

int
get_english_char_width(guchar c)
{
	int idx;
	idx = get_glyph_index(c);
	TT_Load_Glyph(instance, glyph, idx, flags);
	TT_Get_Glyph_Metrics(glyph, &metrics);
	return (metrics.advanceWidth>>6)/xdelta;
}

int 
get_string_width(guchar *text)
{
	int width = 0;

	while(*text) {
		if(is_hangul_font && (*text & 0x80)) {
			width += hangul_font_width/xdelta;
			text += 2;
		} else {
			width += get_english_char_width(*text);
			text++;
		}
		if(*text) width += VALS.spacing;
	}

	return width;
}

FILE *
open_config_file(void)
{
	FILE *fp;
	char config_file[256];
	static char *home = NULL;
	if(!home) home = getenv("HOME");

#if HANGUL
	sprintf(config_file, "%s/.gimp/hangul4gimp.conf", home);
#else
	sprintf(config_file, "%s/.gimp/truetype4gimp.conf", home);
#endif
	fp = fopen(config_file, "r");
	if(fp == NULL) {
		/* no private config file, so search system-wide config file */
		fp = fopen(CONFIG_FILE, "r");
	}
	if(fp == NULL) {
		gdk_beep();
		printf("Can't find the configuration file for TrueType fonts\n");
		return NULL;
	}
	return fp;
}

char **
get_font_list(int *num)
{
	FILE *fp;
	static char **fontnames = NULL;
	static int nlist = 0;
	char buf[256];

	if(fontnames) {
		/* return old values */
		if(num) *num = nlist;
		return fontnames;
	}

	fp = open_config_file();
	if(fp == NULL) {
		if(num) *num = 0;
		return NULL;
	}

	while(fgets(buf, 255, fp) != NULL) {
		if(buf[0] != '#') {
			nlist ++;
		}
	}
	rewind(fp);
	fontnames = g_new(char *, nlist);
	nlist = 0;
	while(fgets(buf, 255, fp) != NULL) {
		if(buf[0] == '#') continue;
		fontnames[nlist++] = g_strdup(strtok(buf, " \t"));
	}
	if(num) *num = nlist;
	return fontnames;
}


char *
get_font_filename(char *fontname)
{
	FILE *fp;
	char fullpath[256];
	char *buf;
	char *font_filename;
	int len;

	fp = open_config_file();
	if(fp == NULL) return NULL;

	encoding = UNI_ENCODING;
	buf = fullpath;
	len = strlen(fontname);
	while(fgets(buf, 255, fp) != NULL) {
		if(strncmp(buf, fontname, len) == 0 && isspace(buf[len])) {
			buf += len;
			font_filename = strchr(buf, '"') + 1; 
			buf = strchr(font_filename, '"');
			*buf = 0;
			font_filename = g_strdup(font_filename);
#if HANGUL
			sscanf(buf + 1, "%d %d %d", &font_index, &is_hangul_font, &encoding);
#endif
			return font_filename;
		}
	}
	return NULL;
}

void
get_cmap(void)
{
	TT_Long i;

    for (i = 0; i < TT_Get_CharMap_Count(face); i++) {
        if (!TT_Get_CharMap_ID(face, i, &pid, &eid)) break;
    }
	if (i == TT_Get_CharMap_Count(face)) nocmap = 1;
	else {
		/* load charmap */
        if (TT_Get_CharMap(face, i, &cmap))
          nocmap = 1;
        else
          nocmap = 0;
    }
}


#if HANGUL
extern guint16 ksc2uni_table[];
guint16 
ksc2uni(guint16 ks)
{
	guint16 page, idx;
	page = (ks >> 8) & 0xff;
	if((page < (33 + 128)) || (page > (125 + 128))) return 0; 
	idx = (ks & 0xff);
	if((idx < 33 + 128) || (idx > 126 + 128)) return 0; 
	return ksc2uni_table[(page - 33 - 128) * (127 - 33) + idx - 33 - 128];
}
#endif

guint16
get_glyph_index(guint16 code)
{
	if(nocmap) return code;
#if HANGUL
	if(code & 0xff00) {
		if(encoding == UNI_ENCODING) {
			return TT_Char_Index(cmap, ksc2uni(code));
		} else {
			if(pid == 3 && eid == 5) return code & 0x7f7f;
			return TT_Char_Index(cmap, code);
		}
	}
#endif
	return TT_Char_Index(cmap, code);
}
/* end of hangul.c */
